using System;
using System.ComponentModel;
using System.Text;

namespace InfiniTec.Configuration.Install
{
    /// <summary>
    /// Represents a field object of an msi-record.
    /// </summary>
    /// <remarks>
    /// This Object uses unmanaged ressources and is therefore disposable.
    /// </remarks>
    public class Field
    {
        private readonly int _FieldNumber;
        private readonly Record _Record;

        /// <summary>
        /// Creates a new instance of the <see cref="Field"/> class.
        /// </summary>
        /// <param name="record">The record this field is part of.</param>
        /// <param name="fieldNumber">The ordinal of this <see cref="Field"/> in the <see cref="Record"/></param>
        internal Field(Record record, int fieldNumber)
        {
            if (record == null) throw new ArgumentNullException("record");

            _Record = record;

            _FieldNumber = fieldNumber;
        }

        /// <summary>
        /// Returns the integer value from a record field. 
        /// </summary>
        public int GetInt32()
        {
            return NativeMethods.MsiRecordGetInteger(_Record.Handle, _FieldNumber);
        }

        /// <summary>
        /// Returns the string value of a record field. 
        /// </summary>
        public string GetString()
        {
            int size = 1024;
            StringBuilder value = new StringBuilder(size);

            int result = NativeMethods.MsiRecordGetString(_Record.Handle, _FieldNumber, value, ref size);

            if (result != 0)
            {
                throw new Win32Exception(result);
            }

            return value.ToString();
        }

        /// <summary>
        /// Reports whether a record field is null. 
        /// </summary>
        public bool IsNull()
        {
            return NativeMethods.MsiRecordIsNull(_Record.Handle, _FieldNumber);
        }

        /// <summary>
        /// Reads bytes from a record stream field into a buffer. 
        /// </summary>
        public byte[] ReadStream()
        {
            byte[] buffer = new byte[DataSize()];
            int size = buffer.Length;

            int result = NativeMethods.MsiRecordReadStream(_Record.Handle, _FieldNumber, buffer, ref size);

            if (result != 0) throw new Win32Exception(result);

            return buffer;
        }

        /// <summary>
        /// Returns the length of a record field. 
        /// </summary>
        public int DataSize()
        {
            return NativeMethods.MsiRecordDataSize(_Record.Handle, _FieldNumber);
        }

        /// <summary>
        /// Sets a record field to an integer field. 
        /// </summary>
        /// <param name="value">The value to be set.</param>
        public void SetInteger(int value)
        {
            int result = NativeMethods.MsiRecordSetInteger(_Record.Handle, _FieldNumber, value);
            if (result != 0)
            {
                throw new Win32Exception(result);
            }
        }

        /// <summary>
        /// Sets a record field to a string field. 
        /// </summary>
        /// <param name="value">The value to be set.</param>
        public void SetString(string value)
        {
            int result = NativeMethods.MsiRecordSetString(_Record.Handle, _FieldNumber, value);
            if (result != 0)
            {
                throw new Win32Exception(result);
            }
        }

        /// <summary>
        /// Sets a record stream field from a file.  
        /// </summary>
        /// <param name="filePath">The path to the file to be streamed.</param>
        public void SetStream(string filePath)
        {
            int result = NativeMethods.MsiRecordSetStream(_Record.Handle, _FieldNumber, filePath);
            if (result != 0)
            {
                throw new Win32Exception(result);
            }
        }
    }
}