using System;
using System.ComponentModel;
using System.Text;
using System.Windows.Forms;

namespace InfiniTec.Configuration.Install
{
    /// <summary>
    /// The msi installer context. 
    /// </summary>
    /// <remarks>
    /// This Object uses unmanaged ressources and is therefore disposable.
    /// </remarks>
    public class InstallerContext : IDisposable
    {
        private static InstallerContext _Current;

        private readonly MsiHandle _Handle;
        private MsiHandle _DatabaseHandle;

        private bool _Disposed;
        private string _CustomActionType;
        private InstallPhase _Phase;

        /// <summary>
        /// Creates a new instance of the <see cref="InstallerContext"/> class.
        /// </summary>
        /// <param name="handle">The handle of the <see cref="InstallerContext"/>.</param>
        public InstallerContext(MsiHandle handle)
        {
            _Handle = handle;
        }

        public static InstallerContext Current
        {
            get { return _Current; }
            internal set { _Current = value; }
        }

        /// <summary>
        /// The <see cref="IntPtr"/> handle of the <see cref="InstallerContext"/>.
        /// </summary>
        public MsiHandle CurrentHandle
        {
            get
            {
                ThrowIfDisposed();
                return _Handle;
            }
        }

     
        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        internal void UpdateTypeContext(InstallPhase phase, Type type)
        {
            if (type == null) throw new ArgumentNullException("type");

            _Phase = phase;
            _CustomActionType = type.FullName;
        }

        public void StartAction(string action, string description, string actionDataFormat)
        {
            if (action == null) throw new ArgumentNullException("action");
            if (description == null) throw new ArgumentNullException("description");
            if (actionDataFormat == null) throw new ArgumentNullException("actionDataFormat");

            using (Record record = Record.Create(3))
            {
                record[1].SetString(action);
                record[2].SetString(description);
                record[3].SetString(actionDataFormat);
                ProcessMessage(InstallMessageType.ActionStart, record);
            }

            using (Record record = Record.Create(3))
            {
                record[1].SetInteger(1);
                record[2].SetInteger(1);
                record[3].SetInteger(0);
                ProcessMessage(InstallMessageType.Progress, record);
            }
        }

        public void LogMessage(string message, params object[] args)
        {
            if (args != null && args.Length > 0)
            {
                message = string.Format(message, args);
            }

            using (Record record = Record.Create(4))
            {
                record[0].SetString("CustomAction ([1], [2]) [3]: [4]");
                record[1].SetString(_Phase.ToString());
                record[2].SetString(_CustomActionType);
                record[3].SetString(DateTime.Now.ToString("[HH:mm:ss:fff]"));
                record[4].SetString(message);

                ProcessMessage(InstallMessageType.Info, record);
            }
        }

        /// <summary>
        /// Reports details about the current action started using the <see cref="StartAction"/> method.
        /// </summary>
        /// <param name="values">The values for the message specified in the <see cref="StartAction"/> method.</param>
        public void ReportDetails(params string[] values)
        {
            if (values == null) throw new ArgumentNullException("values");

            using (Record record = Record.Create(values.Length))
            {
                for (int i = 0; i < values.Length; i++)
                {
                    record[i + 1].SetString(values[i]);
                }
                ProcessMessage(InstallMessageType.ActionData, record);
            }
            using (Record record = Record.Create(3))
            {
                record[1].SetInteger(2);
                record[2].SetInteger(10000);
                record[3].SetInteger(0);
                ProcessMessage(InstallMessageType.Progress, record);
            }
        }

        public DialogResult ShowDialog(string messageTemplate, InstallMessageType messageType, MessageBoxButtons buttons, MessageBoxIcon icon,
                                       params string[] args)
        {
            using (Record record = Record.Create(args.Length + 1))
            {
                record[0].SetString(messageTemplate);
                for (int i = 0; i < args.Length; i++)
                {
                    record[1 + i].SetString(args[i]);
                }

                return (DialogResult) NativeMethods.MsiProcessMessage(_Handle, (uint) messageType | (uint) buttons | (uint) icon, record.Handle);
            }
        }

        public DialogResult ShowDialog(int errorCode, InstallMessageType messageType, MessageBoxButtons buttons, MessageBoxIcon icon,
                                     params string[] args)
        {
            using (Record record = Record.Create(args.Length + 2))
            {
                record[1].SetInteger(errorCode);
                for (int i = 0; i < args.Length; i++)
                {
                    record[2 + i].SetString(args[i]);
                }

                return (DialogResult)NativeMethods.MsiProcessMessage(_Handle, (uint)messageType | (uint)buttons | (uint)icon, record.Handle);
            }
        }


        public void ProcessMessage(InstallMessageType messageType, Record record)
        {
            NativeMethods.MsiProcessMessage(_Handle, (uint) messageType, record.Handle);
        }

        /// <summary>
        /// Executes a built-in action, custom action, or user-interface wizard action. 
        /// </summary>
        /// <param name="action">The name of the action to execute.</param>
        public void DoAction(string action)
        {
            ThrowIfDisposed();


            if (string.IsNullOrEmpty(action))
            {
                throw new ArgumentNullException("action");
            }


            int result = NativeMethods.MsiDoAction(_Handle, action);

            if (result != 0)
            {
                throw new Win32Exception(result);
            }
        }

        /// <summary>
        /// Evaluates a conditional expression containing property names and values. 
        /// </summary>
        /// <param name="conditionId">The id of the condition to evaluate.</param>
        public MsiCondition EvaluateCondition(string conditionId)
        {
            ThrowIfDisposed();

            if (string.IsNullOrEmpty(conditionId))
            {
                throw new ArgumentNullException("conditionId");
            }

            return NativeMethods.MsiEvaluateCondition(_Handle, conditionId);
        }

        /// <summary>
        /// Returns the numeric language of the installation that is currently running. 
        /// </summary>
        public short GetLanguage()
        {
            ThrowIfDisposed();

            return NativeMethods.MsiGetLanguage(_Handle);
        }

        /// <summary>
        /// Gets the value for an installer property. 
        /// </summary>
        /// <param name="propertyName">The id of the property.</param>
        public string GetProperty(string propertyName)
        {
            ThrowIfDisposed();

            if (string.IsNullOrEmpty(propertyName))
            {
                throw new ArgumentNullException("propertyName");
            }

            int size = 1024;
            StringBuilder value = new StringBuilder(size);

            int result = NativeMethods.MsiGetProperty(_Handle, propertyName, value, ref size);

            if (result != 0)
                throw new Win32Exception(result);

            return value.ToString();
        }

        /// <summary>
        /// Returns the full source path for a folder in the Directory table. 
        /// </summary>
        /// <param name="folderId">The id of the folder.</param>
        public string GetSourcePath(string folderId)
        {
            ThrowIfDisposed();

            if (string.IsNullOrEmpty(folderId))
            {
                throw new ArgumentNullException("folderId");
            }

            int size = 1024;
            StringBuilder value = new StringBuilder(size);

            int result = NativeMethods.MsiGetSourcePath(_Handle, folderId, value, ref size);

            if (result != 0)
            {
                throw new Win32Exception(result);
            }

            return value.ToString();
        }

        /// <summary>
        /// Returns the full target path for a folder in the Directory table. 
        /// </summary>
        /// <param name="folderId">The id of the folder.</param>
        public string GetTargetPath(string folderId)
        {
            ThrowIfDisposed();

            if (string.IsNullOrEmpty(folderId))
            {
                throw new ArgumentNullException("folderId");
            }

            int size = 1024;
            StringBuilder value = new StringBuilder(size);

            int result = NativeMethods.MsiGetTargetPath(_Handle, folderId, value, ref size);

            if (result != 0)
            {
                throw new Win32Exception(result);
            }

            return value.ToString();
        }

        /// <summary>
        /// Returns the full path to a Windows Installer component containing an assembly. 
        /// The function prompts for a source and performs any necessary installation. 
        /// ProvideAssembly also increments the usage count for the feature. 
        /// </summary>
        /// <remarks>
        /// This function is available starting with Windows Installer version 2.0
        /// </remarks>
        public string ProvideAssembly(string assemblyName, string installPath, MsiAssemblyInfo assemblyInfo)
        {
            ThrowIfDisposed();

            if (string.IsNullOrEmpty(assemblyName))
            {
                throw new ArgumentNullException("assemblyName");
            }
            if (string.IsNullOrEmpty(installPath))
            {
                throw new ArgumentNullException("installPath");
            }

            int size = 1024;
            StringBuilder value = new StringBuilder(size);

            int result =
                NativeMethods.MsiProvideAssembly(assemblyName, installPath, 0, (int) assemblyInfo, value, ref size);

            if (result == 0)
            {
                return value.ToString();
            }

            return null;
        }

        /// <summary>
        /// Sets the installation level for a full product installation. 
        /// </summary>
        /// <param name="installLevel">The level to set.</param>
        public void SetInstallLevel(int installLevel)
        {
            ThrowIfDisposed();

            int result = NativeMethods.MsiSetInstallLevel(_Handle, installLevel);

            if (result != 0)
            {
                throw new Win32Exception(result);
            }
        }

        /// <summary>
        /// Sets the value for an installation property. 
        /// </summary>
        /// <param name="propertyName">The id of the property to be set.</param>
        /// <param name="value">The value to be set.</param>
        public void SetProperty(string propertyName, string value)
        {
            ThrowIfDisposed();

            if (string.IsNullOrEmpty(propertyName))
            {
                throw new ArgumentNullException("propertyName");
            }
            if (string.IsNullOrEmpty(value))
            {
                throw new ArgumentNullException("value");
            }

            int result = NativeMethods.MsiSetProperty(_Handle, propertyName, value);

            if (result != 0)
            {
                throw new Win32Exception(result);
            }
        }

        /// <summary>
        /// Sets the full target path for a folder in the Directory table. 
        /// </summary>
        /// <param name="folderId">The id of the folder.</param>
        /// <param name="path">The path to set.</param>
        public void SetTargetPath(string folderId, string path)
        {
            ThrowIfDisposed();

            if (string.IsNullOrEmpty(folderId))
            {
                throw new ArgumentNullException("folderId");
            }
            if (string.IsNullOrEmpty(path))
            {
                throw new ArgumentNullException("path");
            }

            int result = NativeMethods.MsiSetTargetPath(_Handle, folderId, path);

            if (result != 0)
            {
                throw new Win32Exception(result);
            }
        }

        
        /// <summary>
        /// Returns the requested state of a feature
        /// </summary>
        /// <param name="feature">The id of the feature to get the state for.</param>
        /// <param name="installed">Specifies the returned current installed <see cref="InstallState"/>.</param>
        /// <param name="action">Receives the  <see cref="InstallState"/> action taken during the installation session.</param>
        public void GetFeatureState(string feature, out InstallState installed, out InstallState action)
        {
            ThrowIfDisposed();

            if (string.IsNullOrEmpty(feature))
            {
                throw new ArgumentNullException("feature");
            }

            int result = NativeMethods.MsiGetFeatureState(_Handle, feature, out installed, out action);
            if (result != 0)
            {
                throw new Win32Exception(result);
            }
        }

        /// <summary>
        /// Returns the requested state of a component
        /// </summary>
        public void GetComponentState(string component, out InstallState installed, out InstallState action)
        {
            ThrowIfDisposed();

            if (string.IsNullOrEmpty(component))
            {
                throw new ArgumentNullException("component");
            }

            int result = NativeMethods.MsiGetComponentState(_Handle, component, out installed, out action);
            if (result != 0)
            {
                throw new Win32Exception(result);
            }
        }

        /// <summary>
        /// Sets a component to the requested state
        /// </summary>
        public void SetComponentState(string component, InstallState installed)
        {
            ThrowIfDisposed();


            if (string.IsNullOrEmpty(component))
            {
                throw new ArgumentNullException("component");
            }


            int result = NativeMethods.MsiSetComponentState(_Handle, component, installed);
            if (result != 0)
            {
                throw new Win32Exception(result);
            }
        }

        /// <summary>
        /// Sets a component to the requested state
        /// </summary>
        public void SetFeatureState(string feature, InstallState installed)
        {
            ThrowIfDisposed();

            if (string.IsNullOrEmpty(feature))
            {
                throw new ArgumentNullException("feature");
            }

            int result = NativeMethods.MsiSetFeatureState(_Handle, feature, installed);
            if (result != 0)
            {
                throw new Win32Exception(result);
            }
        }

        /// <summary>
        /// Returns the error record that was last returned for the calling process.
        /// </summary>
        public Record GetLastError()
        {
            ThrowIfDisposed();

            MsiHandle handle = NativeMethods.MsiGetLastErrorRecord();
            if (!handle.IsInvalid)
            {
                return new Record(handle);
            }

            return null;
        }

        /// <summary>
        /// Prepares a database query and returns a view object
        /// </summary>
        public View OpenView(string query)
        {
            ThrowIfDisposed();

            if (string.IsNullOrEmpty(query))
            {
                throw new ArgumentNullException("query");
            }

            MsiHandle viewHandle;

            _DatabaseHandle = NativeMethods.MsiGetActiveDatabase(_Handle);

            int result = NativeMethods.MsiDatabaseOpenView(_DatabaseHandle, query, out viewHandle);

            if (result != 0)
            {
                throw new Win32Exception(result);
            }

            return new View(viewHandle);
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        /// <param name="disposing"><see langword="true"/> if the object is currently disposing; otherwise, <see langword="false"/>.</param>
        protected virtual void Dispose(bool disposing)
        {
            ThrowIfDisposed();

            _Disposed = true; 
            
            if (disposing)
            {
                if (_DatabaseHandle != null)
                {
                    _DatabaseHandle.Dispose();
                    _DatabaseHandle = null;
                }
            }

        }

        /// <summary>
        /// Throws an <see cref="ObjectDisposedException"/> if this instance is already disposed.
        /// </summary>
        protected void ThrowIfDisposed()
        {
            if (_Disposed)
            {
                throw new ObjectDisposedException(GetType().Name);
            }
        }

        public bool GetMode(InstallerRunMode mode)
        {
            return NativeMethods.MsiGetMode(_Handle, mode);
        }
    }
}