using System;
using System.Collections;
using System.ComponentModel;

namespace InfiniTec.Configuration.Install
{
    public class Record : CollectionBase, IDisposable
    {
        private readonly MsiHandle _Handle;

        internal Record(MsiHandle handle)
        {
            if (handle == null) throw new ArgumentNullException("handle");

            if (handle.IsInvalid)
            {
                throw new ArgumentNullException("handle");
            }

            int fieldCount = NativeMethods.MsiRecordGetFieldCount(handle);
            for (int index = 0; index <= fieldCount; index++)
            {
                Add(new Field(this, index));
            }

            _Handle = handle;
        }

        private Record(MsiHandle handle, int fieldCount)
        {
            if (handle == null) throw new ArgumentNullException("handle");

            if (handle.IsInvalid)
            {
                throw new ArgumentException("The specified handle is invalid", "handle");
            }

            for (int index = 0; index <= fieldCount; index++)
            {
                Add(new Field(this, index));
            }

            _Handle = handle;
        }

        /// <summary>
        /// The <see cref="IntPtr"/> handle of the record.
        /// </summary>
        internal MsiHandle Handle
        {
            get
            {
                ThrowIfDisposed();
                return _Handle;
            }
        }

        /// <summary>
        /// The <see cref="Field"/>list.
        /// </summary>
        /// <param name="index">The ordinal of the <see cref="Field"/>.</param>
        /// <returns>The specified <see cref="Field"/>.</returns>
        public Field this[int index]
        {
            get
            {
                ThrowIfDisposed();
                return (Field)InnerList[index];
            }
            set
            {
                ThrowIfDisposed();
                InnerList[index] = value;
            }
        }

        public static Record Create(int columnCount)
        {
            MsiHandle handle = NativeMethods.MsiCreateRecord(columnCount);

            return new Record(handle, columnCount);
        }

        /// <summary>
        /// Sets all fields in a record to null.
        /// </summary>
        public void ClearData()
        {
            ThrowIfDisposed();

            int result = NativeMethods.MsiRecordClearData(_Handle);
            if (result != 0)
            {
                throw new Win32Exception(result);
            }
        }


        private void Add(Field value)
        {
            ThrowIfDisposed();

            if (value == null)
            {
                throw new ArgumentNullException("value");
            }

            InnerList.Add(value);
        }

        private bool _Disposed;

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        /// <param name="disposing"><see langword="true"/> if the object is currently disposing; otherwise, <see langword="false"/>.</param>
        protected virtual void Dispose(bool disposing)
        {
            ThrowIfDisposed();

            _Disposed = true;

            if (disposing)
            {
                _Handle.Dispose();
            }

        }

        /// <summary>
        /// Throws an <see cref="ObjectDisposedException"/> if this instance is already disposed.
        /// </summary>
        protected void ThrowIfDisposed()
        {
            if (_Disposed)
            {
                throw new ObjectDisposedException(GetType().Name);
            }
        }
    }
}