using System;
using System.Collections;
using System.ComponentModel;
using System.Configuration.Install;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using System.Runtime.Serialization.Formatters.Binary;
using System.Windows.Forms;
using System.Xml.Serialization;
using System.Collections.Generic;

namespace InfiniTec.Configuration.Install
{
	/// <summary>
	/// Execute managed custom actions from dynamically loaded assemblies.
	/// </summary>
	/// <remarks>
	/// This Object uses unmanaged ressources and is therefore disposable.
	/// </remarks>
	public class Runtime
	{
		/// <summary>
		/// Entry point for running managed custom actions.
		/// </summary>
		/// <param name="handle">The handle of the current msi instance.</param>
        public int RunActions(long handle)
		{
            using (InstallerLogTextWriter consoleLogWriter = new InstallerLogTextWriter())
            using (InstallerLogTextWriter traceLogWriter = new InstallerLogTextWriter())
            {
                try
                {
                    MsiHandle msiHandle = new MsiHandle(new IntPtr(handle));
                    using (InstallerContext context = new InstallerContext(msiHandle))
                    {
                        InstallerContext.Current = context;

                        Console.SetOut(consoleLogWriter);
                        Trace.Listeners.Add(new TextWriterTraceListener(traceLogWriter));


                        Trace.WriteLine("CustomActionRuntime running...");

                        string configurationXml;
                        if (context.GetMode(InstallerRunMode.Scheduled))
                        {
                            configurationXml = context.GetProperty("CustomActionData");
                        }
                        else
                        {
                            configurationXml = context.GetProperty("ManagedCustomActionConfiguration");
                        }

                        SimpleCustomActionConfiguration configuration = DeserializeConfiguration<SimpleCustomActionConfiguration>(configurationXml);

                        Assembly assembly = GetAssemblyToRun(configuration);

                        ICustomAction customAction;

                        Type type = assembly.GetType(configuration.TypeName, true);

                        InstallerContext.Current.UpdateTypeContext(InstallPhase.CustomAction, type);

                        customAction = Activator.CreateInstance(type) as ICustomAction;
                        if (customAction != null)
                        {
                            customAction.Execute();
                        }

                        return 0;
                    }
                }
                catch (Exception ex)
                {
                    InstallerContext.Current.LogMessage(ex.ToString());
                    return -3;
                }
            }
		}

	    public int ManagedInstall(long handle)
        {
            return RunInstaller(handle, InstallPhase.Install);
        }
        public int ManagedUninstall(long handle)
        {
            return RunInstaller(handle, InstallPhase.Uninstall);
        }
        public int ManagedRollback(long handle)
        {
            return RunInstaller(handle, InstallPhase.Rollback);
        }
        public int ManagedCommit(long handle)
        {
            return RunInstaller(handle, InstallPhase.Commit);
        }

	    private static int RunInstaller(long handle, InstallPhase phase)
	    {
            using (InstallerLogTextWriter consoleLogWriter = new InstallerLogTextWriter())
            using (InstallerLogTextWriter traceLogWriter = new InstallerLogTextWriter())
            {
                try
                {
                    MsiHandle msiHandle = new MsiHandle(new IntPtr(handle));

                    using (InstallerContext context = new InstallerContext(msiHandle))
                    {
                        InstallerContext.Current = context;

                        Console.SetOut(consoleLogWriter);
                        Trace.Listeners.Add(new TextWriterTraceListener(traceLogWriter));

                        Trace.WriteLine(string.Format("Running {0} phase of local assembly.", phase));

                        string configurationXml = context.GetProperty("CustomActionData");

                        CustomActionConfiguration configuration = DeserializeConfiguration<CustomActionConfiguration>(configurationXml);

                        Assembly assembly = LoadCustomActionFromFileSystem(configuration);

                        try
                        {
                            IDictionary stateSaver = GetStateSaver(configuration, phase);
                            RunInstallers(assembly, phase, stateSaver, configuration.Properties);

                            if (phase != InstallPhase.Uninstall)
                            {
                                SaveSavedState(stateSaver, configuration);
                            }
                        }
                        finally
                        {
                            if (phase == InstallPhase.Uninstall)
                            {
                                DeleteSavedState(configuration);
                            }
                        }

                        return 0;
                    }
                }
                catch (Exception ex)
                {
                    Trace.WriteLine(ex.ToString());
                    InstallerContext.Current.ShowDialog((int) ((phase == InstallPhase.Uninstall) ? ErrorCode.CustomActionUninstallError : ErrorCode.CustomActionError),
                                                        InstallMessageType.Error, MessageBoxButtons.OK, MessageBoxIcon.Error,
                                                        ex.Message);
                    InstallerContext.Current.LogMessage(ex.ToString());
                    return -3;
                }
            }
	    }

	    private static void DeleteSavedState(CustomActionConfiguration configuration)
	    {
	        string filename = GetStateSaverFilename(configuration);

            if (File.Exists(filename))
            {
                File.Delete(filename);
            }
	    }

	    private static IDictionary GetStateSaver(CustomActionConfiguration configuration, InstallPhase phase)
	    {
	        if (configuration == null) throw new ArgumentNullException("configuration");

	        if (phase == InstallPhase.Install)
	        {
	            return new Hashtable();
	        }
            else
	        {
	            string stateSaverFile = GetStateSaverFilename(configuration);

	            BinaryFormatter serializer = new BinaryFormatter();

                if (!File.Exists(stateSaverFile)) return new Hashtable();

                using (Stream stream = File.OpenRead(stateSaverFile))
                {
                    return (Hashtable) serializer.Deserialize(stream);
                }
	        }
	    }

	    private static void SaveSavedState(IDictionary savedState, CustomActionConfiguration configuration)
        {
            if (savedState == null) throw new ArgumentNullException("savedState");
            if (configuration == null) throw new ArgumentNullException("configuration");

            string stateSaverFile = GetStateSaverFilename(configuration);

            BinaryFormatter serializer = new BinaryFormatter();

            using (Stream stream = File.Create(stateSaverFile))
            {
                serializer.Serialize(stream, savedState);
            }   
        }

	    private static string GetStateSaverFilename(CustomActionConfiguration configuration)
	    {
	        return Path.Combine(Path.GetDirectoryName(configuration.File),
	                            Path.GetFileName(configuration.File) + ".installstate");
	    }

	    private static void RunInstallers(Assembly assembly, InstallPhase phase, IDictionary stateSaver, IEnumerable<Parameter> parameters)
	    {
	        foreach (Type type in assembly.GetTypes())
	        {
                Trace.WriteLine(string.Format("Checking wheter type {0} is an enabled managed installer class...", type.FullName));
	            if (typeof(Installer).IsAssignableFrom(type) && InstallerIsEnabled(type))
	            {
	                using (Installer instance = (Installer) Activator.CreateInstance(type))
	                {
                        InstallerContext.Current.UpdateTypeContext(phase, type);

	                    instance.Context = new InstallContext();

	                    foreach (Parameter parameter in parameters)
	                    {
	                        instance.Context.Parameters.Add(parameter.Name, parameter.Value);
	                    }

                        switch (phase)
                        {
                            case InstallPhase.Install:
                                instance.Install(stateSaver);
                                break;
                            case InstallPhase.Commit:
                                instance.Commit(stateSaver);
                                break;
                            case InstallPhase.Rollback:
                                instance.Rollback(stateSaver);
                                break;
                            case InstallPhase.Uninstall:
                                instance.Uninstall(stateSaver);
                                break;
                            default:
                                break;
                        }
	                }
	            }
	        }
	    }

	    private static bool InstallerIsEnabled(Type type)
	    {
	        RunInstallerAttribute attribute = Attribute.GetCustomAttribute(type, typeof (RunInstallerAttribute)) as RunInstallerAttribute;

	        return (attribute == null) ? false : attribute.RunInstaller;
	    }

	    private static T DeserializeConfiguration<T>(string configurationXml)
            where T: CustomActionConfiguration
	    {
	        Debug.WriteLine("ConfigurationXml: " + configurationXml);
	        T configuration;

	        XmlSerializer serializer = new XmlSerializer(typeof(T));

	        configuration = (T) serializer.Deserialize(new StringReader(configurationXml));
	        return configuration;
	    }

	    private static Assembly GetAssemblyToRun(SimpleCustomActionConfiguration configuration)
	    {
	        Debug.WriteLine("CustomActionAssembly: " + configuration.TypeName);

	        Assembly assembly;
	        if (!string.IsNullOrEmpty(configuration.File))
	        {
	            assembly = LoadCustomActionFromFileSystem(configuration);
	        }
	        else
	        {
	            assembly = LoadInstallerAssemblyFromBinaryTable(configuration);
	        }
	        return assembly;
	    }

	    private static Assembly LoadCustomActionFromFileSystem(CustomActionConfiguration configuration)
        {
            Trace.WriteLine(string.Format("Loading assembly {0}", configuration.File));
            Assembly assembly = Assembly.LoadFrom(configuration.File);

            return assembly;
        }

        private static Assembly LoadInstallerAssemblyFromBinaryTable(SimpleCustomActionConfiguration configuration)
        {
            string sql = string.Format("SELECT * FROM Binary WHERE Name = '{0}'", configuration.BinaryKey);
            using (View view = InstallerContext.Current.OpenView(sql))
            {
                RecordCollection records = view.Execute();

                if (records.Count == 0)
                {
                    Trace.WriteLine(string.Format("Could not find assembly with binary key {0}", configuration.BinaryKey));
                }

                Assembly assembly = Assembly.Load(records[0][2].ReadStream());

                return assembly;
            }
        }       
	}
}
