using System;
using System.ComponentModel;

namespace InfiniTec.Configuration.Install
{
    public class View : IDisposable
    {
        private const long ERROR_NO_MORE_ITEMS = 259L;
        private bool _Disposed;

        private MsiHandle _Handle;


        /// <summary>
        /// Creates a new instance of the 
        /// </summary>
        /// <param name="handle">The handle of the view.</param>
        internal View(MsiHandle handle)
        {
            _Handle = handle;
        }


        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Executes a SQL view query and supplies any required parameters.
        /// </summary>
        public RecordCollection Execute(Record parameters)
        {
            ThrowIfDisposed();

            int result;
            if (parameters != null)
                result = NativeMethods.MsiViewExecute(_Handle, parameters.Handle);
            else
                result = NativeMethods.MsiViewExecute(_Handle, MsiHandle.Zero);

            if (result != 0)
            {
                throw new Win32Exception(result);
            }
            else
            {
                RecordCollection records = new RecordCollection();

                for (;;)
                {
                    MsiHandle recordHandle;
                    result = NativeMethods.MsiViewFetch(_Handle, out recordHandle);
                    if (result == ERROR_NO_MORE_ITEMS)
                    {
                        break;
                    }
                    records.Add(new Record(recordHandle));
                }

                return records;
            }
        }

        /// <summary>
        /// Executes a SQL view query and supplies any required parameters.
        /// </summary>
        public RecordCollection Execute()
        {
            ThrowIfDisposed();

            return Execute(null);
        }

        /// <summary>
        /// Releases the result set for an executed view.
        /// </summary>
        public void Close()
        {
            ThrowIfDisposed();

            NativeMethods.MsiViewClose(_Handle);
            Dispose();
        }

        /// <summary>
        /// Finalizes this instance.
        /// </summary>
        ~View()
        {
            if (!_Disposed)
            {
                Dispose(false);
            }
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        /// <param name="disposing"><see langword="true"/> if the object is currently disposing; otherwise, <see langword="false"/>.</param>
        protected virtual void Dispose(bool disposing)
        {
            ThrowIfDisposed();

            if (disposing)
            {
                if (_Handle != null)
                {
                    _Handle.Dispose();
                    _Handle = null;
                }
            }
            _Disposed = true;
        }

        /// <summary>
        /// Throws an <see cref="ObjectDisposedException"/> if this instance is already disposed.
        /// </summary>
        protected void ThrowIfDisposed()
        {
            if (_Disposed)
            {
                throw new ObjectDisposedException(GetType().Name);
            }
        }
    }
}