using System;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Xml;
using System.Xml.Schema;
using System.Xml.Serialization;
using InfiniTec.Configuration.Install;
using Microsoft.Tools.WindowsInstallerXml;
using Microsoft.Tools.WindowsInstallerXml.Serialize;
using Row=Microsoft.Tools.WindowsInstallerXml.Row;
using YesNoType=Microsoft.Tools.WindowsInstallerXml.YesNoType;

namespace ManagedInstallerWixExtension
{
    internal class ManagedInstallerCompiler : CompilerExtension
    {
        private readonly XmlSchema _Schema = LoadXmlSchemaHelper(Assembly.GetExecutingAssembly(), "ManagedInstallerWixExtension.Include.ManagedCustomActionSchema.xsd");

        public override void ParseElement(SourceLineNumberCollection sourceLineNumbers, XmlElement parentElement,
                                          XmlElement element, params string[] contextValues)
        {
            switch (element.LocalName)
            {
                case "ManagedCustomAction":
                    ParseManagedCustomActionElement(element);
                    break;
                case "ManagedActionSequence":
                    ParseManagedActionSequenceElement(element);
                    break;
                case "ManagedInstaller":
                    ParseManagedInstallerElement(element, parentElement, sourceLineNumbers);
                    break;
                default:
                    Core.UnexpectedElement(parentElement, element);
                    break;
            }
        }

        private void ParseManagedInstallerElement(XmlNode element, XmlElement parentElement, SourceLineNumberCollection sourceLineNumbers)
        {
            XmlElement component = (XmlElement) parentElement.ParentNode;
            if (component.LocalName != "Component")
            {
                return;
            }

            List<Parameter> parameters = new List<Parameter>();
            foreach (XmlNode node in element.ChildNodes)
            {
                if (node.NodeType != XmlNodeType.Element)
                {
                    continue;
                }

                if (node.LocalName == "Parameter")
                {
                    parameters.Add(ParseParameterElement((XmlElement) node));
                }
                else
                {
                    Core.UnexpectedElement(element, node);
                }
            }

            CreateAndScheduleManagedInstaller(parentElement.GetAttribute("Id"), 
                component.GetAttribute("Id"), parameters, sourceLineNumbers);
        }

        private static Parameter ParseParameterElement(XmlElement element)
        {
            Parameter result = new Parameter();
            result.Name = element.GetAttribute("Name");
            result.Value = element.InnerText;

            return result;
        }

        private void CreateAndScheduleManagedInstaller(string fileKey, string componentKey, 
            List<Parameter> parameters, SourceLineNumberCollection sourceLineNumbers)
        {
            CreateCustomActionRow(fileKey + ".Install", (int) CustomActionType.Dll | (int) CustomActionSequence.InScript,
                "CLRHosting", "ManagedInstall", sourceLineNumbers);
            CreateCustomActionRow(fileKey + ".Uninstall", (int)CustomActionType.Dll | (int)CustomActionSequence.InScript
                | (int) CustomActionExecutionOptions.IgnoreReturnStatus,

                "CLRHosting", "ManagedUninstall", sourceLineNumbers);
            CreateCustomActionRow(fileKey + ".Rollback", (int)CustomActionType.Dll | (int)CustomActionSequence.InScript |
                (int) CustomActionSequence.Rollback, "CLRHosting", "ManagedRollback", sourceLineNumbers);
            CreateCustomActionRow(fileKey + ".Commit", (int)CustomActionType.Dll | (int)CustomActionSequence.InScript |
                (int)CustomActionSequence.Commit, "CLRHosting", "ManagedCommit", sourceLineNumbers);

            SequenceCustomAction(new CustomActionSettings(fileKey + ".Install", "InstallServices", null, "InstallExecuteSequence", string.Format("${0}>2", componentKey), sourceLineNumbers));
            SequenceCustomAction(new CustomActionSettings(fileKey + ".Commit", "InstallServices", null, "InstallExecuteSequence", string.Format("${0}>2", componentKey), sourceLineNumbers));
            SequenceCustomAction(new CustomActionSettings(fileKey + ".Rollback", "InstallServices", null, "InstallExecuteSequence", string.Format("${0}>2", componentKey), sourceLineNumbers));
            SequenceCustomAction(new CustomActionSettings(fileKey + ".Uninstall", "UnpublishComponents", null, "InstallExecuteSequence", string.Format("${0}=2", componentKey), sourceLineNumbers));

            CustomActionConfiguration config = new CustomActionConfiguration();
            config.File = string.Format("[#{0}]", fileKey);
            config.Properties = parameters;

            string configValue = SerializeConfig(config);

            CreateCustomActionRow(fileKey + ".Install.Configuration", (int)CustomActionType.TextData | (int)CustomActionSource.Property, fileKey + ".Install", configValue, sourceLineNumbers);
            CreateCustomActionRow(fileKey + ".Uninstall.Configuration", (int)CustomActionType.TextData | (int)CustomActionSource.Property, fileKey + ".Uninstall", configValue, sourceLineNumbers);
            CreateCustomActionRow(fileKey + ".Commit.Configuration", (int)CustomActionType.TextData | (int)CustomActionSource.Property, fileKey + ".Commit", configValue, sourceLineNumbers);
            CreateCustomActionRow(fileKey + ".Rollback.Configuration", (int)CustomActionType.TextData | (int)CustomActionSource.Property, fileKey + ".Rollback", configValue, sourceLineNumbers);

            SequenceCustomAction(new CustomActionSettings(fileKey + ".Install.Configuration", fileKey + ".Install", null, "InstallExecuteSequence", string.Format("${0}>2", componentKey), sourceLineNumbers));
            SequenceCustomAction(new CustomActionSettings(fileKey + ".Rollback.Configuration", fileKey + ".Rollback", null, "InstallExecuteSequence", string.Format("${0}>2", componentKey), sourceLineNumbers));
            SequenceCustomAction(new CustomActionSettings(fileKey + ".Commit.Configuration", fileKey + ".Commit", null, "InstallExecuteSequence", string.Format("${0}>2", componentKey), sourceLineNumbers));
            SequenceCustomAction(new CustomActionSettings(fileKey + ".Uninstall.Configuration", fileKey + ".Uninstall", null, "InstallExecuteSequence", string.Format("${0}=2", componentKey), sourceLineNumbers));
        }

        private void ParseManagedActionSequenceElement(XmlNode element)
        {
            foreach (XmlNode node in element.ChildNodes)
            {
                if (node.NodeType != XmlNodeType.Element || node.LocalName != "Managed") continue;

                ParseManagedElement((XmlElement) node);
            }
            
        }

        private void ParseManagedElement(XmlElement node)
        {
            Custom custom;
            custom = new Custom();
            string table;
            int sequenceValue;
            int? sequence;
            string actionName;

            actionName = node.GetAttribute("Action");

            
            if (int.TryParse(node.GetAttribute("Sequence"), out sequenceValue))
            {
                sequence = sequenceValue;
            }
            else
            {
                sequence = null;
            }

            SourceLineNumberCollection childSourceLineNumbers = Preprocessor.GetSourceLineNumbers(node);

            table = node.GetAttribute("SequenceTable");
            string condition = CompilerCore.GetConditionInnerText(node);
            SequenceCustomAction(new CustomActionSettings(actionName, node.GetAttribute("Before"), node.GetAttribute("After"), sequence, table, condition, childSourceLineNumbers));

            custom.Before = actionName;
            custom.Action = actionName + ".Configuration";

            SequenceCustomAction(new CustomActionSettings(actionName + ".Configuration", actionName, node.GetAttribute("After"), sequence, table, condition, childSourceLineNumbers));
        }


        private void SequenceCustomAction(CustomActionSettings customActionSettings)
        {
            Row row = Core.CreateRow(customActionSettings.ChildSourceLineNumbers, "WixAction");

            ReferenceWixLib(customActionSettings.ChildSourceLineNumbers);

            row[0] = customActionSettings.Table;
            row[1] = customActionSettings.Action;
            row[2] = customActionSettings.Condition;
            if (customActionSettings.Sequence != null)
            {
                row[3] = customActionSettings.Sequence;
            }
            row[4] = customActionSettings.Before;
            row[5] = customActionSettings.After;
            row[6] = 0;
        }

        private void ReferenceWixLib(SourceLineNumberCollection childSourceLineNumbers)
        {
            Core.CreateWixSimpleReferenceRow(childSourceLineNumbers, "Binary", "CLRHosting");
            Core.CreateWixSimpleReferenceRow(childSourceLineNumbers, "Binary", "CARuntime");
            Core.CreateWixSimpleReferenceRow(childSourceLineNumbers, "Binary", "MsiHostingConfig");
            Core.CreateWixSimpleReferenceRow(childSourceLineNumbers, "CustomAction", "ManagedInitialize");
//            Core.CreateWixSimpleReferenceRow(childSourceLineNumbers, "Error", "28000");
//            Core.CreateWixSimpleReferenceRow(childSourceLineNumbers, "Error", "28001");
        }

        private void ParseManagedCustomActionElement(XmlElement element)
        {
            string id;
            SimpleCustomActionConfiguration config;

            config = new SimpleCustomActionConfiguration();
            id = element.GetAttribute("Id");

            config.TypeName= element.GetAttribute("Type");
            config.BinaryKey= element.GetAttribute("BinaryKey");
            config.File = string.Format("[#{0}]", element.GetAttribute("FileKey"));

            SourceLineNumberCollection childSourceLineNumbers = Preprocessor.GetSourceLineNumbers(element);

            Core.CreateWixSimpleReferenceRow(childSourceLineNumbers, "Binary", "CLRHosting");

            int executionMode = GetExecutionMode(element, childSourceLineNumbers);

            string result = SerializeConfig(config);

            CreateCustomActionRow(id, executionMode, "CLRHosting", "Execute", childSourceLineNumbers);

            string targetProperty;

            if (((CustomActionSequence) executionMode & CustomActionSequence.InScript) == CustomActionSequence.InScript)
            {
                targetProperty = id;    
            }
            else
            {
                targetProperty = "ManagedCustomActionConfiguration";
            }
            CreateCustomActionRow(id + ".Configuration", (int) CustomActionType.TextData | (int) CustomActionSource.Property, targetProperty, result, childSourceLineNumbers);
        }

        private static string SerializeConfig(CustomActionConfiguration config)
        {
            if (config == null) throw new ArgumentNullException("config");

            XmlSerializer serializer = new XmlSerializer(config.GetType());
            StringWriter writer = new StringWriter();

            serializer.Serialize(writer, config);
            return writer.ToString();
        }

        private void CreateCustomActionRow(string id, int executionMode, string target, string value, SourceLineNumberCollection childSourceLineNumbers)
        {
            Row row = Core.CreateRow(childSourceLineNumbers, "CustomAction");
            row[0] = id;        
            row[1] = executionMode;
            row[2] = target;
            row[3] = value;
        }

      

        private int GetExecutionMode(XmlElement element, SourceLineNumberCollection sourceLineNumbers)
        {
            int bits = (int) CustomActionType.Dll | (int) CustomActionSource.BinaryData;
            
            if (GetAttributeYesNoValue(sourceLineNumbers, element.GetAttributeNode("HideTarget")) == YesNoType.Yes)
            {
                bits |= (int) CustomActionExecutionOptions.HideTarget;
            }

            if (GetAttributeYesNoValue(sourceLineNumbers, element.GetAttributeNode("Impersonate")) == YesNoType.No)
            {
                bits |= (int) CustomActionExecutionOptions.NoImpersonate;
            }

            if (GetAttributeYesNoValue(sourceLineNumbers, element.GetAttributeNode("PatchUninstall")) == YesNoType.Yes)
            {
                bits |= (int)CustomActionSequence.PatchUninstall;
            }

            switch (element.GetAttribute("Return"))
            {
                case CompilerCore.IllegalEmptyAttributeValue: // this error is already handled
                    break;
                case "asyncNoWait":
                    bits |= (int) CustomActionExecutionOptions.RunAsync | (int) CustomActionExecutionOptions.IgnoreReturnStatus;
                    break;
                case "asyncWait":
                    bits |= (int) CustomActionExecutionOptions.RunAsync;
                    break;
                case "check":
                    break;
                case "ignore":
                    bits |= (int) CustomActionExecutionOptions.IgnoreReturnStatus;
                    break;
                default:
                    Core.OnMessage(WixErrors.IllegalAttributeValue(sourceLineNumbers, element.Name, element.Name, element.GetAttribute("Return"), "asyncNoWait", "asyncWait", "check", "ignore"));
                    break;
            }

            switch (element.GetAttribute("Execute"))
            {
                case CompilerCore.IllegalEmptyAttributeValue: // this error is already handled
                    break;
                case "commit":
                    bits |= (int) CustomActionSequence.InScript | (int) CustomActionSequence.Commit;
                    break;
                case "deferred":
                    bits |= (int) CustomActionSequence.InScript;
                    break;
                case "firstSequence":
                    bits |= (int)CustomActionSequence.FirstSequence;
                    break;
                case "immediate":
                    break;
                case "oncePerProcess":
                    bits |= (int)CustomActionSequence.OncePerProcess;
                    break;
                case "rollback":
                    bits |= (int) CustomActionSequence.InScript | (int) CustomActionSequence.OncePerProcess;
                    break;
                case "secondSequence":
                    bits |= (int) CustomActionSequence.ClientRepeat;
                    break;
                default:
                    Core.OnMessage(WixErrors.IllegalAttributeValue(sourceLineNumbers, element.Name, element.Name, element.GetAttribute("Execute"), "commit", "deferred", "firstSequence", "immediate", "oncePerProcess", "rollback", "secondSequence"));
                    break;
            }

            return bits;
        }

        private YesNoType GetAttributeYesNoValue(SourceLineNumberCollection numbers, XmlAttribute attribute)
        {
            if (attribute == null)
            {
                return YesNoType.NotSet;
            }
            return Core.GetAttributeYesNoValue(numbers, attribute);
        }

        public override XmlSchema Schema
        {
            get { return _Schema; }
        }
    }
}